#!/usr/bin/python3
import os
import sys

if 'SUMO_HOME' in os.environ:
     tools = os.path.join(os.environ['SUMO_HOME'], 'tools')
     sys.path.append(tools)
else:
     sys.exit("please declare environment variable 'SUMO_HOME'")

# import time
from sumolib import checkBinary
from enum import Enum
import traceback
import argparse

parser = argparse.ArgumentParser(
    prog='runner.py',
    #usage='%(prog)s [optional arguments] -- SUMOARGS',
    description='Run SUMO and compute passengers per vehicle',
    epilog="Examples\n"
          '  runner.py -t 5000 \n'
          '    run SUMO until 5000 sec'
, formatter_class=argparse.RawDescriptionHelpFormatter)
parser.add_argument('-f','--first', type=int, default=3600, help='From time')
parser.add_argument('-l','--last', type=int, default=60000, help='last time step for simulation')
parser.add_argument('-s','--seed', type=int, default=42, help='random seed')
parser.add_argument('-g','--gui', action='store_true', help='Use sumo-gui')
parser.add_argument('-L','--libsumo', action='store_true', help='Use libsumo')
parser.add_argument('-F','--fcd', action='store_true', help='export fcd data')
args = parser.parse_args()

'''
call tree
  run  -+->  fueling_behavior
        +->  maintain_headway  -> maintainHeadway
        +->  takeDriverBreak   -> insertDriverBreak (not yet)

Functions for common tasks
  extendCurrentStop
  printVehicleMessage
  printVehicleStateMessage
'''

if args.libsumo:
  import libsumo as traci
else:
  import traci

class BusState(Enum):
  running = 0
  goingToCharge = 1
  charging = 2
  waitingForGoingToCharge = 3

class DriverBreakStatus(Enum):
    running = 0
    AMbathroomFood = 5
    PMbathroomFood = 6
    shiftChange = 7
    Lunch = 8
#     notBreaking = 8

breakStates = {"0":0, "1":0, "2":0, "3":0, "4":0, "5":0, "6":0, "7":0}
vehicles = ("0", "1", "2", "3", "4", "5", "6", "7") # This was previously a list
finishedBuses = [] 
activeBuses = []
waitingBuses = {}
chargingStationEdge = '-7855681#1'
# have a dictionary of vehicles and their states, initially all running
busStates = {}

csvfile = None
energyfile = None
eventfile = None

def fuelingHeadway(vehID, headway):
    traci.vehicle.setMinGap(vehID, headway)

def printVehicleStateMessage(vehicleId, message, states = busStates):
    global eventfile

    stateMessage = '(%s) %s' % (states[vehicleId].name, message)
    printVehicleMessage(vehicleId, stateMessage)

def printVehicleMessage(vehicleId, message):
    global eventfile
    global current_time

    longMessage = 'time %5ds vehicle %s %s' % (current_time, vehicleId, message)
    print(longMessage)
    print(longMessage, file=eventfile, flush=True)

# plugins for the main loop fror different questions
def fueling_behavior(current_time):
  global csvfile, energyfile, eventfile
  global actualFuelLevel, numActualFuelLevel, numBusRiders
  global activeBuses

  if 3600 <= current_time <= 65000:

    simTime = [str(current_time)]
    for e in activeBuses:
        try:
          actualFuelLevel = traci.vehicle.getParameter(e, "device.battery.actualBatteryCapacity")
        except Exception as ex:
          if str(ex).endswith(' is not known.'):
            printVehicleStateMessage(e, 'error %s' % ( str(ex)))
            finishedBuses.append(e)
            activeBuses.remove(e)
            printVehicleStateMessage(e, 'finishedBuses %s' % ( str(finishedBuses)))
            printVehicleStateMessage(e, 'e in finishedBuses %s' % (e in finishedBuses))
          else:
            traceback.print_exc(file=eventfile)
          continue
        numActualFuelLevel = float(actualFuelLevel)
        numBusRiders, stopState = traci.vehicle.getPersonNumber(e), traci.vehicle.getStopState(e)
        # here is the state machine
        if busStates[e] == BusState.running:
          fuelingHeadway(e, 1) # if the vehicle is ABs; OTHERWISE COMMENT OUT <<<<----
          roadID = traci.vehicle.getRoadID(e)
          if numActualFuelLevel <= float(140000) and numBusRiders == 0 and (not roadID.startswith(':')): #666667:Diesel/Biodiesel, 665102: Propane/NaturalGas, 86400:Electricity, 212400:Hydrogen (40% fuel depletion)
            printVehicleStateMessage(e, 'send to charging station (fuel %10.2f, stopState %d)' % (numActualFuelLevel, stopState))
            try:
                nextStop = 0
                if stopState != 0:
                  nextStop = 1
                traci.vehicle.replaceStop(e, nextStop, "chargingStation_-7855681#1_0_0", duration=300, flags=32) #chargingStation_-7855681#1_0_0 # duration=900 (HDBs), ABs-duration Diesel/Biodiesel:600, Propane/NaturalGas:600, Electricity:3000, Hydrogen:300
                busStates[e], routeIndex, route = BusState.goingToCharge, traci.vehicle.getRouteIndex(e), traci.vehicle.getRoute(e)
                if routeIndex >= 0:
                  try:
                    newStopIndex = route.index(chargingStationEdge, routeIndex)
                  except:
                    newStopIndex = -1
                  printVehicleStateMessage(e, 'New route from index %d to %d' % ( routeIndex, newStopIndex))
                #print('new route for vehicle %s = %s' % (e, ' '.join(traci.vehicle.getRoute(e))))
            except Exception as ex:
                traceback.print_exc(file=eventfile)
                #print(str(e))
                #print ("Unexpected error:", sys.exc_info()[0])
                printVehicleStateMessage(e, 'error cannot set new stop (%s)' % (str(ex)))
                busStates[e], roadID = BusState.waitingForGoingToCharge, traci.vehicle.getRoadID(e)
                printVehicleStateMessage(e, 'is at %s' % (roadID))
                # wait 100 sec and try again
                waitingBuses[e] = 100
        elif busStates[e] == BusState.waitingForGoingToCharge:
          waitingBuses[e] -= 1
          if waitingBuses[e] == 0:
            busStates[e] = BusState.running
            printVehicleStateMessage(e, 'is running again')
        elif busStates[e] == BusState.goingToCharge:
          roadID = traci.vehicle.getRoadID(e)
          if roadID == chargingStationEdge:
            fuelingHeadway(e, 8) # if the vehicle is ABs; OTHERWISE COMMENT OUT <<<<----
            busStates[e] = BusState.charging
            printVehicleStateMessage(e, 'at %-30s fuel %10.2f' % (roadID, numActualFuelLevel))
        elif busStates[e] == BusState.charging:
          roadID = traci.vehicle.getRoadID(e)
          #printVehicleStateMessage(e, 'at %-30s fuel %10.2f' % (roadID, numActualFuelLevel))
          # charging stops, when
          #   the stop at the charging station is finished
          #   or a resume command is issued
          if roadID != chargingStationEdge:
            busStates[e] = BusState.running
          if numActualFuelLevel > 221666: #1055555:Diesel/Biodiesel, 1053078:Propane/NaturalGas, 136800:Electricity, 336300:Hydrogen (95% full)
            try:
              printVehicleStateMessage(e, 'resume (fuel %10.2f)' % (numActualFuelLevel))
              traci.vehicle.resume(e)
              busStates[e] = BusState.running
            except Exception as ex:
                printVehicleStateMessage(e, 'error cannot resume (%s)' % (str(ex)))
            actualFuelLevel = 0
        simTime.append(str(actualFuelLevel))
        simTime.append(str(numBusRiders))
    #print('fuel and pass', ' '.join(simTime))
    print(','.join(simTime), file=energyfile, flush=True)

#       print ('%.0f Vehicle %s Battery %8.2f' % (current_time, '0', float(current_fuel_Level)))
def extendCurrentStop(vehID, newDuration, stopData = None):
    '''Common routine to extend stop duration'''
    try:
        if stopData == None: # --- WHY IS THIS NEEDED? ARE YOU TRYING TO CONTROL ERRORS IN THE CODE? ---
            stopData = traci.vehicle.getStops(vehID, 1)
        traci.vehicle.setBusStop(vehID, stopData[0].stoppingPlaceID,
                                 duration=newDuration)
        newStopData = traci.vehicle.getStops(vehID, 1)
        #printVehicleMessage(vehID, 'new %s' % newStopData)
        if newStopData[0].duration != newDuration:
          printVehicleMessage(vehID, 'setting of duration does not work')
    except Exception as ex:
        printVehicleMessage(vehID, 'error with extendCurrentStop %s' % (str(ex)))
        printVehicleMessage(vehID, 'stopData old %s' % (stopData[0]))

def insertDriverBreak(vehId, stops, duration, reason, newDriverBreakStatus):
  '''Construction area - one routine for all types of break'''
  stopData = traci.vehicle.getStops(vehId, 1)
  #print(stopData)
  if (stopData[0].stopFlags & 8) and stopData[0].stoppingPlaceID in stops: # if bus is at a bus stop and bus stop is one of stops in break list
    printVehicleMessage(vehId, 'is taking a %d min %s break' %
                        (duration/60, reason))
    extendCurrentStop(vehId, duration, stopData)
    breakStates[vehId] = newDriverBreakStatus

def takeDriverBreak(current_time):
    global DriverStops, currentDriverStop
    stopsForAMbreak = ["busStop_-467228692#3_1_10", "busStop_JagTranRdSB7_1_8",
                       "busStop_GreekRowSB20_1_1", "busStop_GreekRowNB4_1_0",
                       "Y1"
      ]
    stopsForLunchbreak = ["busStop_-467228692#3_1_10", "busStop_JagTranRdSB7_1_8",
                          "busStop_GreekRowSB20_1_1", "busStop_GreekRowNB4_1_0"
      ]
    stopsForPMbreak = ["busStop_-467228692#3_1_10", "busStop_JagTranRdSB7_1_8",
                          "busStop_GreekRowSB20_1_1", "busStop_GreekRowNB4_1_0"
      ]

    for e in activeBuses:
        if (10800 < current_time < 12600):
            if (breakStates[e] != DriverBreakStatus.AMbathroomFood):
                if busStates[e] == BusState.running and numBusRiders == 0 and numActualFuelLevel > float(666667):
                    insertDriverBreak(e, stopsForAMbreak, 600, 'AM Bathroom/Snack', DriverBreakStatus.AMbathroomFood) # duration=300 for HBD
        if (21600 < current_time < 27000):
            if (breakStates[e] != DriverBreakStatus.Lunch):
                if busStates[e] == BusState.running and numBusRiders == 0 and numActualFuelLevel > float(666667):
                    insertDriverBreak(e, stopsForLunchbreak, 600, 'Lunch', DriverBreakStatus.Lunch) # duration=300 for HBD
        if (30600 < current_time < 37800):
            if (breakStates[e] != DriverBreakStatus.PMbathroomFood):
                if busStates[e] == BusState.running and numBusRiders == 0 and numActualFuelLevel > float(666667):
                    insertDriverBreak(e, stopsForPMbreak, 600, 'PM Bathroom/Snack', DriverBreakStatus.PMbathroomFood) # duration=300 for HBD
#         if 30600 < current_time < 37800:
#             if busStates[e] == BusState.running and numBusRiders == 0 and numActualFuelLevel > float(416625):
#                 if currentDriverStop == "busStop_-467228692#3_1_10" or currentDriverStop == "busStop_JagTranRdSB7_1_8" or currentDriverStop == "busStop_GreekRowSB20_1_1" or currentDriverStop == "busStop_GreekRowNB4_1_0" or currentDriverStop == "Y1":
#                     try:
#                         print("Time: %7.2fs, Vehicle %s is taking a 30 min Lunch Break" % (current_time, e))
#                         traci.vehicle.setBusStop(e, currentDriverStop, duration=1800, flags=0)
#                         breakStates[e] = DriverBreakStatus.Lunch
#                     except:
#                         print("Time: %7.2fs, Vehicle %s does not need a Lunch Break" % (current_time, e))

def maintain_headway(current_time):
    #Maintaining headway between Yellow Buses, Blue Buses, & Green Buses; respectively
    #bus_ids = ["0", "1", "2", "3", "4", "5", "6", "7"]
    simulation_time = [str(current_time)]
    if current_time >= 3600:
        maintainHeadway('3', '4', delay=140) #HDBdelay=200, ABdelay=140, originally 160
        maintainHeadway('4', '3', delay=140) #HDBdelay=200, ABdelay=140 
        maintainHeadway('1', '2', delay=240) #HDBdelay=300, ABdelay=240
        maintainHeadway('2', '1', delay=240) #HDBdelay=300, ABdelay=240
        maintainHeadway('5', '6', delay=100) #HDBdelay=160, ABdelay=100
        maintainHeadway('6', '5', delay=100) #HDBdelay=160, ABdelay=100

def maintainHeadway(leaderID, followerID, delay=160): #this could be set to 100 secs
    global finishedBuses

    if leaderID in finishedBuses or followerID in finishedBuses:
      return
    try:
        # First check if leader is at a bus stop (for performance)
        leaderIsAtBusStop = traci.vehicle.isAtBusStop(leaderID)
        #printVehicleMessage(leaderID, 'is at bus stop %s' % (leaderIsAtBusStop))
        if leaderIsAtBusStop: # --- I THOUGHT YOU NEEDED TO VERIFY THIS AGAINST A TRUE, SINCE IT RETURNS A BOOLEAN? ---
          followersStopData = traci.vehicle.getStops(followerID, 2)
          #printVehicleMessage(followerID, 'followers next stop %s' % (followersStopData))
          if len(followersStopData) > 0 and followersStopData[0].duration != delay:
            leaderStopData = traci.vehicle.getStops(leaderID, 1)
            leadersCurrentStop = leaderStopData[0].stoppingPlaceID
            leadersCurrentLane   = leaderStopData[0].lane #stoppingPlaceID for CURRENT
            followersCurrentLane = followersStopData[0].lane #stoppingPlaceID for CURRENT
            followersCurrentLane1 = followersStopData[1].lane # --- IS THIS HOW YOU LOOK AT THE LANEID AT NEXT STOP AHEAD? ---
            followersCurrentStop = followersStopData[0].stoppingPlaceID #stoppingPlaceID for CURRENT
            followersCurrentStop1 = followersStopData[1].stoppingPlaceID # --- IS THIS HOW YOU LOOK AT THE NEXT STOP AHEAD? ---
            #printVehicleMessage(leaderID,   ' leader   current stop %s' % (leadersCurrentLane))
            #printVehicleMessage(followerID, ' follower current stop %s)' % (followersCurrentLane))
#             if (leadersCurrentLane == followersCurrentLane) or (leadersCurrentLane == followersCurrentLane1):
            if (leadersCurrentStop == followersCurrentStop) or (leadersCurrentStop == followersCurrentStop1):
                if (leadersCurrentLane == followersCurrentLane) or (leadersCurrentLane == followersCurrentLane1):
                  printVehicleMessage(followerID, 'waits at %s for %d sec (headway)' % (followersCurrentStop, delay))
                  extendCurrentStop(followerID, delay, followersStopData)
                  #checkStatus = traci.vehicle.getStops(followerID, 1)
                  #printVehicleMessage(followerID, 'new duration %d' % (checkStatus[0].duration))
    except Exception as ex:
      print('Vehicle %s error %s' % (followerID, str(ex)))
      traceback.print_exc(file=eventfile)

def run(seed, gui = False, endTime = 60000):
  global busStates, breakStates
  global csvfile, energyfile, eventfile
  global current_time
  global finishedBuses
  global activeBuses

  # define file names
  csvfile = open('passengers.%d.out.csv' % seed, 'w')
  energyfile = open("batteryCapacity.%d.out.csv" % seed, "w")
  eventfile = open("events.%d.out.txt" % seed, "w")
  chargingOutput = "chargingstations.%d.out.xml" % seed
  stopOutput = "stop.%d.out.xml" % seed
  batteryOutput = "battery.%d.out.xml" % seed
  personOutput = "person.%d.out.xml" % seed

  if gui:
    sumoBinary = checkBinary ("sumo-gui")
  else:
    sumoBinary = checkBinary ("sumo")

  sumoCmd = [sumoBinary, "--quit-on-end", "--start"]
  sumoCmd.extend(['--seed', str(seed), '-c', 'test.sumocfg'])
  sumoCmd.extend(['--chargingstations-output', chargingOutput])
  sumoCmd.extend(['--stop-output', stopOutput])
  sumoCmd.extend(['--battery-output', batteryOutput])
  sumoCmd.extend(['--tripinfo-output', personOutput, '--device.tripinfo.probability', '0'])
  if args.fcd:
    sumoCmd.extend(['--fcd-output', 'fcd.%s.out.xml' % seed, '--person-device.fcd.probability', '0', '--fcd-output.distance'])

  print('Run ' + ' '.join(sumoCmd))

  # reset global mvariables for each run
  activeBuses = []
  for v in vehicles:
    busStates[v] = BusState.running
    breakStates[v] = DriverBreakStatus.running
    activeBuses.append(v)
  finishedBuses = []
  waitingBuses = {}

  traci.start(sumoCmd)
  step = 0

  # main loop, current_time serves as step
  while traci.simulation.getMinExpectedNumber() > 0:
    traci.simulationStep()
    current_time = traci.simulation.getTime()
    if current_time <= 1:
      printVehicleMessage('0', 'start simulation')

    # Execution for Beginning of Simulation
    # do nothing special
    # Execution for collecting Ridership within Bus Vehicles
    # stop if all buses are gone
    if len(activeBuses) == 0:
      break
    # call plugins
    fueling_behavior(current_time)
    maintain_headway(current_time)
#     takeDriverBreak(current_time)
    if 3600 <= step:
          # This set of code gets the number of riders currently within the bus
        list_of_ridership = [str(current_time)]
        for id in vehicles:
          # take care, if vehicle does not exist
          ridership = 0
          if not id in finishedBuses:
            try:
              ridership = traci.vehicle.getPersonNumber(id)
              vehicleDistance = traci.vehicle.getDistance(id)
              vehicleEnergyCons = traci.vehicle.getElectricityConsumption(id)
            except:
              pass
          list_of_ridership.append(str(ridership))
          list_of_ridership.append(str(vehicleDistance))
          list_of_ridership.append(str(vehicleEnergyCons))
          list_of_ridership.append(str(actualFuelLevel))
        # create a comma-separated list
        print(','.join(list_of_ridership), file=csvfile, flush=True)
    if current_time > endTime:
      break
    step += 1 # not really used any more

  csvfile.close()
  eventfile.close()
  traci.close()

if __name__ == "__main__":
  run(args.seed, args.gui, args.last)

# ===============================PARKING LOT============================================================================================
#       current_fuel_Level_Red = traci.vehicle.getParameter(vehicle[0], "device.battery.actualBatteryCapacity")
#       current_fuel_Level_Blue1 = traci.vehicle.getParameter(vehicle[1], "device.battery.actualBatteryCapacity")
#       current_fuel_Level_Blue2 = traci.vehicle.getParameter(vehicle[2], "device.battery.actualBatteryCapacity")
#       current_fuel_Level_Yellow1 = traci.vehicle.getParameter(vehicle[3], "device.battery.actualBatteryCapacity")
#       current_fuel_Level_Yellow2 = traci.vehicle.getParameter(vehicle[4], "device.battery.actualBatteryCapacity")
#       current_fuel_Level_Green1 = traci.vehicle.getParameter(vehicle[5], "device.battery.actualBatteryCapacity")
#       current_fuel_Level_Green2 = traci.vehicle.getParameter(vehicle[6], "device.battery.actualBatteryCapacity")
#       current_fuel_Level_Orange = traci.vehicle.getParameter(vehicle[7], "device.battery.actualBatteryCapacity")
#       print("At timestep %8.2f: %8.2f, %8.2f, %8.2f, %8.2f, %8.2f, %8.2f, %8.2f, %8.2f" % (current_time, float(current_fuel_Level_Red), float(current_fuel_Level_Blue1),
#                                                                         float(current_fuel_Level_Blue2), float(current_fuel_Level_Yellow1), float(current_fuel_Level_Yellow2),
#                                                                         float(current_fuel_Level_Green1), float(current_fuel_Level_Green2), float(current_fuel_Level_Orange)))

        # Turning Lane ID into Edge ID to use as variable for setting Stop for bus 3
#         detect_EdgeID = str(traci.vehicle.getLaneID("3"))
#         edge_ID = detect_EdgeID[:-2] #detect_EdgeID.replace("_1","")

#For Green Route Buses (ONLY)
  # Detect Vehicle 6 so that Vehicle 5 can slow down
#   =========================================================
#   if current_time >= 3600:
#     secondVehicle = '3'
#     leader = traci.vehicle.getLeader(secondVehicle, 5)
#     atBusStop = traci.vehicle.isAtBusStop(secondVehicle)
#     if (atBusStop != False) and leader != None:
#         print("At timestep %.0f JagTran %s is at Bus Stop" % (current_time, secondVehicle))
#         print('leader', leader)
#         #print('Leader of 5 %s' % leader_id)
#         if leader[0] == "4" and leader[1] <= 10:
#              oldSpeed = traci.vehicle.getSpeed(secondVehicle)
#              traci.vehicle.slowDown(secondVehicle, 0.02, 65)
#              print('%.0f Slowdown vehicle %s with current speed %4.1f' % (current_time, secondVehicle, oldSpeed))
# ==========================================================================

#              traci.vehicle.getNextStops(secondVehicle)
#              traci.vehicle.replaceStop(secondVehicle, 1, "GreekRowSB20", 12.10, 1, 7.15, 460, 0)
#             traci.vehicle.setSpeed(secondVehicle, 0.00)

#         if leader_id >= ("6", 15):
#             traci.vehicle.openGap("5", 360, 1850, 300, -0.5, -0.5, "6")
#             traci.vehicle.getSt
